from dagster import Field, Shape


def logging_handler_schema() -> Shape:
    return Shape(
        description="Provide configuration for the OpenTelemetry logging handler.",
        fields={
            "level": Field(
                str,
                description="The minimum logging level to capture.",
                is_required=False,
            ),
        },
    )



from dagster import Array, Field, Permissive, StringSource

SHARED_DOCKER_CONFIG = {
    "networks": Field(
        Array(StringSource),
        is_required=False,
        description=(
            "Names of the networks to which to connect the launched container at creation time"
        ),
    ),
    "env_vars": Field(
        [str],
        is_required=False,
        description=(
            "The list of environment variables names to include in the docker container. "
            "Each can be of the form KEY=VALUE or just KEY (in which case the value will be pulled "
            "from the local environment)"
        ),
    ),
    "container_kwargs": Field(
        Permissive(),
        is_required=False,
        description=(
            "key-value pairs that can be passed into containers.create. See "
            "https://docker-py.readthedocs.io/en/stable/containers.html for the full list "
            "of available options."
        ),
    ),
}



import logging
import sys
import time
from typing import Any, NamedTuple, Optional

import docker
import docker.errors
from dagster import (
    Field,
    IntSource,
    _check as check,
)
from dagster._core.launcher.base import LaunchRunContext
from dagster._core.utils import parse_env_var
from dagster._grpc.types import ExecuteRunArgs
from dagster._serdes import ConfigurableClass
from dagster._serdes.config_class import ConfigurableClassData
from dagster._utils import find_free_port
from dagster._utils.merger import merge_dicts
from dagster._vendored.dateutil.parser import parse
from dagster_cloud_cli.core.workspace import PexMetadata
from dagster_docker import DockerRunLauncher
from dagster_docker.container_context import DockerContainerContext
from dagster_shared.serdes.serdes import deserialize_value, serialize_value
from docker.models.containers import Container
from typing_extensions import Self

from dagster_cloud.api.dagster_cloud_api import UserCodeDeploymentType
from dagster_cloud.execution.monitoring import CloudContainerResourceLimits
from dagster_cloud.storage.tags import PEX_METADATA_TAG
from dagster_cloud.workspace.user_code_launcher.user_code_launcher import UserCodeLauncherEntry
from dagster_cloud.workspace.user_code_launcher.utils import get_grpc_server_env

from ..config_schema.docker import SHARED_DOCKER_CONFIG
from ..user_code_launcher import (
    DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT,
    SHARED_USER_CODE_LAUNCHER_CONFIG,
    DagsterCloudGrpcServer,
    DagsterCloudUserCodeLauncher,
    ServerEndpoint,
)
from ..user_code_launcher.utils import deterministic_label_for_location
from .utils import unique_docker_resource_name

GRPC_SERVER_LABEL = "dagster_grpc_server"
MULTIPEX_SERVER_LABEL = "dagster_multipex_server"
AGENT_LABEL = "dagster_agent_id"
SERVER_TIMESTAMP_LABEL = "dagster_server_timestamp"
STOP_TIMEOUT_LABEL = "dagster_stop_timeout"


IMAGE_PULL_LOG_INTERVAL = 15


class DagsterDockerContainer(NamedTuple):
    """We use __str__ on server handles to serialize them to the run tags. Wrap the docker container
    object so that we can serialize it to a string.
    """

    container: Container

    def __str__(self):  # pyright: ignore[reportIncompatibleMethodOverride]
        return self.container.id


class DockerUserCodeLauncher(
    DagsterCloudUserCodeLauncher[DagsterDockerContainer], ConfigurableClass
):
    def __init__(
        self,
        inst_data=None,
        networks=None,
        env_vars=None,
        container_kwargs=None,
        **kwargs,
    ):
        self._inst_data = inst_data
        self._logger = logging.getLogger("dagster_cloud")
        self._networks = check.opt_list_param(networks, "networks", of_type=str)
        self._input_env_vars = check.opt_list_param(env_vars, "env_vars", of_type=str)

        self._container_kwargs = check.opt_dict_param(
            container_kwargs, "container_kwargs", key_type=str
        )

        super().__init__(**kwargs)

    @property
    def requires_images(self):
        return True

    def user_code_deployment_type(self) -> UserCodeDeploymentType:  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        return UserCodeDeploymentType.DOCKER

    @property
    def inst_data(self):
        return self._inst_data

    @property
    def env_vars(self) -> list[str]:
        return self._input_env_vars + self._instance.dagster_cloud_api_env_vars

    @property
    def container_kwargs(self) -> dict[str, Any]:
        return self._container_kwargs

    @classmethod
    def config_type(cls):
        return merge_dicts(
            SHARED_DOCKER_CONFIG,
            {
                "server_process_startup_timeout": Field(
                    IntSource,
                    is_required=False,
                    default_value=DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT,
                    description=(
                        "Timeout when waiting for a code server to be ready after it is created"
                    ),
                ),
            },
            SHARED_USER_CODE_LAUNCHER_CONFIG,
        )

    @classmethod
    def from_config_value(cls, inst_data: ConfigurableClassData, config_value: Any) -> Self:
        return cls(inst_data=inst_data, **config_value)

    def _create_container(
        self,
        client,
        image,
        container_name,
        hostname,
        environment,
        ports,
        container_context,
        command,
        labels,
    ):
        container_kwargs = {**container_context.container_kwargs}

        container_kwargs.pop("stop_timeout", None)

        return client.containers.create(
            image,
            detach=True,
            hostname=hostname,
            name=container_name,
            network=container_context.networks[0] if len(container_context.networks) else None,
            environment=environment,
            labels=labels,
            command=command,
            ports=ports,
            **container_kwargs,
        )

    def _get_standalone_dagster_server_handles_for_location(
        self, deployment_name: str, location_name: str
    ) -> list[DagsterDockerContainer]:
        client = docker.client.from_env()
        return [
            DagsterDockerContainer(container=container)
            for container in client.containers.list(
                all=True,
                filters={
                    "label": [
                        GRPC_SERVER_LABEL,
                        deterministic_label_for_location(deployment_name, location_name),
                        f"{AGENT_LABEL}={self._instance.instance_uuid}",
                    ]
                },
            )
        ]

    def _get_multipex_server_handles_for_location(
        self, deployment_name: str, location_name: str
    ) -> list[DagsterDockerContainer]:
        client = docker.client.from_env()
        return [
            DagsterDockerContainer(container=container)
            for container in client.containers.list(
                all=True,
                filters={
                    "label": [
                        MULTIPEX_SERVER_LABEL,
                        deterministic_label_for_location(deployment_name, location_name),
                        f"{AGENT_LABEL}={self._instance.instance_uuid}",
                    ]
                },
            )
        ]

    def _launch_container(
        self,
        deployment_name: str,
        location_name: str,
        container_name: str,
        hostname: str,
        grpc_port: int,
        ports: dict[int, int],
        image: str,
        container_context: DockerContainerContext,
        command: list[str],
        additional_env: dict[str, str],
        labels: dict[str, str],
    ) -> tuple[Container, ServerEndpoint]:
        client = docker.client.from_env()

        self._logger.info(
            f"Starting a new container for {deployment_name}:{location_name} with image {image}:"
            f" {container_name}"
        )

        environment = merge_dicts(
            (dict([parse_env_var(env_var) for env_var in container_context.env_vars])),
            additional_env,
        )

        try:
            container = self._create_container(
                client,
                image,
                container_name,
                hostname,
                environment,
                ports,
                container_context,
                command,
                labels,
            )
        except docker.errors.ImageNotFound:
            last_log_time = time.time()
            self._logger.info(f"Pulling image {image}...")
            for _line in docker.APIClient().pull(image, stream=True):
                if time.time() - last_log_time > IMAGE_PULL_LOG_INTERVAL:
                    self._logger.info(f"Still pulling image {image}...")
                    last_log_time = time.time()

            container = self._create_container(
                client,
                image,
                container_name,
                hostname,
                environment,
                ports,
                container_context,
                command,
                labels,
            )

        if len(container_context.networks) > 1:
            for network_name in container_context.networks[1:]:
                network = client.networks.get(network_name)
                network.connect(container)

        container.start()

        endpoint = ServerEndpoint(
            host=hostname,
            port=grpc_port,
            socket=None,
        )

        self._logger.info(f"Started container {container.id}")

        return (container, endpoint)

    def get_code_server_resource_limits(
        self, deployment_name: str, location_name: str
    ) -> CloudContainerResourceLimits:
        # Empty because we don't currently internally monitor docker deployment resource usage.
        return {}

    def _start_new_server_spinup(
        self,
        deployment_name: str,
        location_name: str,
        desired_entry: UserCodeLauncherEntry,
    ) -> DagsterCloudGrpcServer:
        metadata = desired_entry.code_location_deploy_data
        container_name = unique_docker_resource_name(deployment_name, location_name)

        container_context = DockerContainerContext(
            registry=None,
            env_vars=self.env_vars
            + [f"{k}={v}" for k, v in (metadata.cloud_context_env or {}).items()],
            networks=self._networks,
            container_kwargs=self._container_kwargs,
        ).merge(DockerContainerContext.create_from_config(metadata.container_context))

        ports = {}

        has_network = len(self._networks) > 0
        if has_network:
            grpc_port = 4000
            hostname = container_name
        else:
            grpc_port = find_free_port()
            ports[grpc_port] = grpc_port
            hostname = "localhost"

        if metadata.pex_metadata:
            command = metadata.get_multipex_server_command(
                grpc_port,
                metrics_enabled=self._instance.user_code_launcher.code_server_metrics_enabled,
            )
            environment = metadata.get_multipex_server_env()
            labels = {
                MULTIPEX_SERVER_LABEL: "",
                deterministic_label_for_location(deployment_name, location_name): "",
                AGENT_LABEL: self._instance.instance_uuid,
                SERVER_TIMESTAMP_LABEL: str(desired_entry.update_timestamp),
            }
        else:
            command = metadata.get_grpc_server_command(
                metrics_enabled=self._instance.user_code_launcher.code_server_metrics_enabled
            )
            environment = get_grpc_server_env(
                metadata,
                grpc_port,
                location_name,
                self._instance.ref_for_deployment(deployment_name),
            )
            labels = {
                GRPC_SERVER_LABEL: "",
                deterministic_label_for_location(deployment_name, location_name): "",
                AGENT_LABEL: self._instance.instance_uuid,
                SERVER_TIMESTAMP_LABEL: str(desired_entry.update_timestamp),
            }

        if "stop_timeout" in container_context.container_kwargs:
            labels[STOP_TIMEOUT_LABEL] = str(container_context.container_kwargs["stop_timeout"])

        container, server_endpoint = self._launch_container(
            deployment_name,
            location_name,
            container_name,
            hostname,
            grpc_port,
            ports,
            check.not_none(metadata.image),
            container_context,
            command,
            additional_env=environment,
            labels=labels,
        )

        return DagsterCloudGrpcServer(
            DagsterDockerContainer(container=container), server_endpoint, metadata
        )

    async def _wait_for_new_server_ready(  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        self,
        deployment_name: str,
        location_name: str,
        user_code_launcher_entry: UserCodeLauncherEntry,
        server_handle: DagsterDockerContainer,
        server_endpoint: ServerEndpoint,
    ) -> None:
        await self._wait_for_dagster_server_process(
            client=server_endpoint.create_client(),
            timeout=self._server_process_startup_timeout,
        )

    def _remove_server_handle(self, server_handle: DagsterDockerContainer) -> None:
        container = server_handle.container

        stop_timeout_str = server_handle.container.labels.get(STOP_TIMEOUT_LABEL)
        stop_timeout = int(stop_timeout_str) if stop_timeout_str else None

        try:
            container.stop(timeout=stop_timeout)
        except Exception:
            self._logger.error(f"Failure stopping container {container.id}: {sys.exc_info()}")
        container.remove(force=True)
        self._logger.info(f"Removed container {container.id}")

    def get_agent_id_for_server(self, handle: DagsterDockerContainer) -> Optional[str]:
        return handle.container.labels.get(AGENT_LABEL)

    def get_server_create_timestamp(self, handle: DagsterDockerContainer) -> Optional[float]:
        created_time_str = handle.container.attrs["Created"]
        return parse(created_time_str).timestamp()  # pyright: ignore[reportAttributeAccessIssue]

    def _list_server_handles(self) -> list[DagsterDockerContainer]:
        client = docker.client.from_env()
        return [
            DagsterDockerContainer(container=container)
            for container in client.containers.list(all=True, filters={"label": GRPC_SERVER_LABEL})
        ] + [
            DagsterDockerContainer(container=container)
            for container in client.containers.list(
                all=True, filters={"label": MULTIPEX_SERVER_LABEL}
            )
        ]

    def run_launcher(self):
        launcher = CloudDockerRunLauncher(
            image=None,
            env_vars=self.env_vars,
            networks=self._networks,
            container_kwargs=self._container_kwargs,
        )
        launcher.register_instance(self._instance)

        return launcher


class CloudDockerRunLauncher(DockerRunLauncher):
    def launch_run(self, context: LaunchRunContext) -> None:
        serialized_pex_metadata = context.dagster_run.tags.get(PEX_METADATA_TAG)

        if serialized_pex_metadata:
            run = context.dagster_run
            job_origin = check.not_none(run.job_code_origin)

            docker_image = self._get_docker_image(job_origin)

            run_args = ExecuteRunArgs(
                job_origin=job_origin,
                run_id=run.run_id,
                instance_ref=self._instance.get_ref(),
            )

            deserialize_value(serialized_pex_metadata, PexMetadata)
            command = [
                "dagster-cloud",
                "pex",
                "execute-run",
                serialize_value(run_args),
                serialized_pex_metadata,
            ]
            self._launch_container_with_command(run, docker_image, command)
        else:
            return super().launch_run(context)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
If you make changes to this file, run "python -m dagster_cloud.pex.grpc.compile" after."""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class Empty(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___Empty = Empty

@typing_extensions.final
class CreatePexServerRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREATE_PEX_SERVER_ARGS_FIELD_NUMBER: builtins.int
    create_pex_server_args: builtins.str
    def __init__(
        self,
        *,
        create_pex_server_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["create_pex_server_args", b"create_pex_server_args"]) -> None: ...

global___CreatePexServerRequest = CreatePexServerRequest

@typing_extensions.final
class CreatePexServerReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREATE_PEX_SERVER_RESPONSE_FIELD_NUMBER: builtins.int
    create_pex_server_response: builtins.str
    def __init__(
        self,
        *,
        create_pex_server_response: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["create_pex_server_response", b"create_pex_server_response"]) -> None: ...

global___CreatePexServerReply = CreatePexServerReply

@typing_extensions.final
class GetPexServersRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    GET_PEX_SERVERS_ARGS_FIELD_NUMBER: builtins.int
    get_pex_servers_args: builtins.str
    def __init__(
        self,
        *,
        get_pex_servers_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["get_pex_servers_args", b"get_pex_servers_args"]) -> None: ...

global___GetPexServersRequest = GetPexServersRequest

@typing_extensions.final
class GetPexServersReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    GET_PEX_SERVERS_RESPONSE_FIELD_NUMBER: builtins.int
    get_pex_servers_response: builtins.str
    def __init__(
        self,
        *,
        get_pex_servers_response: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["get_pex_servers_response", b"get_pex_servers_response"]) -> None: ...

global___GetPexServersReply = GetPexServersReply

@typing_extensions.final
class GetCrashedPexServersRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    GET_CRASHED_PEX_SERVERS_ARGS_FIELD_NUMBER: builtins.int
    get_crashed_pex_servers_args: builtins.str
    def __init__(
        self,
        *,
        get_crashed_pex_servers_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["get_crashed_pex_servers_args", b"get_crashed_pex_servers_args"]) -> None: ...

global___GetCrashedPexServersRequest = GetCrashedPexServersRequest

@typing_extensions.final
class GetCrashedPexServersReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    GET_CRASHED_PEX_SERVERS_RESPONSE_FIELD_NUMBER: builtins.int
    get_crashed_pex_servers_response: builtins.str
    def __init__(
        self,
        *,
        get_crashed_pex_servers_response: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["get_crashed_pex_servers_response", b"get_crashed_pex_servers_response"]) -> None: ...

global___GetCrashedPexServersReply = GetCrashedPexServersReply

@typing_extensions.final
class ShutdownPexServerRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SHUTDOWN_PEX_SERVER_ARGS_FIELD_NUMBER: builtins.int
    shutdown_pex_server_args: builtins.str
    def __init__(
        self,
        *,
        shutdown_pex_server_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["shutdown_pex_server_args", b"shutdown_pex_server_args"]) -> None: ...

global___ShutdownPexServerRequest = ShutdownPexServerRequest

@typing_extensions.final
class ShutdownPexServerReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SHUTDOWN_PEX_SERVER_RESPONSE_FIELD_NUMBER: builtins.int
    shutdown_pex_server_response: builtins.str
    def __init__(
        self,
        *,
        shutdown_pex_server_response: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["shutdown_pex_server_response", b"shutdown_pex_server_response"]) -> None: ...

global___ShutdownPexServerReply = ShutdownPexServerReply

@typing_extensions.final
class PingRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ECHO_FIELD_NUMBER: builtins.int
    echo: builtins.str
    def __init__(
        self,
        *,
        echo: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["echo", b"echo"]) -> None: ...

global___PingRequest = PingRequest

@typing_extensions.final
class PingReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ECHO_FIELD_NUMBER: builtins.int
    echo: builtins.str
    def __init__(
        self,
        *,
        echo: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["echo", b"echo"]) -> None: ...

global___PingReply = PingReply



from collections.abc import Generator, Iterable, Sequence
from typing import Callable, Optional, Union, cast

# (Gauge is only exposed as a private class)
from opentelemetry.metrics import (
    CallbackOptions,
    Counter,
    Histogram,
    Instrument,
    ObservableCounter,
    ObservableGauge,
    ObservableUpDownCounter,
    Observation,
    UpDownCounter,
)
from opentelemetry.metrics._internal import Meter as SDKMeter

InstrumentKeyType = tuple[str, str, str]
ObservableCallback = Callable[[CallbackOptions], Iterable[Observation]]
ObservableGenerator = Generator[Iterable[Observation], CallbackOptions, None]
ObserverType = Union[ObservableCallback, ObservableGenerator]


class Meter:
    def __init__(self, meter: SDKMeter):
        self._meter = meter
        self._instruments: dict[InstrumentKeyType, Instrument] = {}

    @property
    def name(self) -> str:
        return self._meter.name

    @property
    def version(self) -> Optional[str]:
        return self._meter.version

    @property
    def schema_url(self) -> Optional[str]:
        return self._meter.schema_url

    def get_counter(self, name: str, description: str = "", unit: str = "") -> Counter:
        key = (name, description, unit)
        instrument = self._instruments.get(key)
        if not instrument:
            instrument = self._meter.create_counter(
                name=name,
                description=description,
                unit=unit,
            )
            self._instruments[key] = instrument

        return cast("Counter", instrument)

    def get_up_down_counter(
        self, name: str, description: str = "", unit: str = ""
    ) -> UpDownCounter:
        key = (name, description, unit)
        instrument = self._instruments.get(key)
        if not instrument:
            instrument = self._meter.create_up_down_counter(
                name=name,
                description=description,
                unit=unit,
            )
            self._instruments[key] = instrument

        return cast("UpDownCounter", instrument)

    def get_histogram(self, name: str, description: str = "", unit: str = "") -> Histogram:
        key = (name, description, unit)
        instrument = self._instruments.get(key)
        if not instrument:
            instrument = self._meter.create_histogram(
                name=name,
                description=description,
                unit=unit,
            )
            self._instruments[key] = instrument

        return cast("Histogram", instrument)

    def get_observable_counter(
        self,
        name: str,
        description: str = "",
        unit: str = "",
        observers: Optional[Sequence[ObserverType]] = None,
    ) -> ObservableCounter:
        key = (name, description, unit)
        instrument = self._instruments.get(key)
        if not instrument:
            instrument = self._meter.create_observable_counter(
                name=name,
                description=description,
                unit=unit,
                callbacks=observers,
            )
            self._instruments[key] = instrument

        return cast("ObservableCounter", instrument)

    def get_observable_up_down_counter(
        self,
        name: str,
        description: str = "",
        unit: str = "",
        observers: Optional[Sequence[ObserverType]] = None,
    ) -> ObservableUpDownCounter:
        key = (name, description, unit)
        instrument = self._instruments.get(key)
        if not instrument:
            instrument = self._meter.create_observable_up_down_counter(
                name=name,
                description=description,
                unit=unit,
                callbacks=observers,
            )
            self._instruments[key] = instrument

        return cast("ObservableUpDownCounter", instrument)

    def get_observable_gauge(
        self,
        name: str,
        description: str = "",
        unit: str = "",
        observers: Optional[Sequence[ObserverType]] = None,
    ) -> ObservableGauge:
        key = (name, description, unit)
        instrument = self._instruments.get(key)
        if not instrument:
            instrument = self._meter.create_observable_gauge(
                name=name,
                description=description,
                unit=unit,
                callbacks=observers,
            )
            self._instruments[key] = instrument

        return cast("ObservableGauge", instrument)



from dagster import BoolSource, Field, Shape, StringSource

from dagster_cloud.opentelemetry.config.exporter import exporter_config_schema
from dagster_cloud.opentelemetry.config.log_record_processor import log_record_processor_schema
from dagster_cloud.opentelemetry.config.logging_handler import logging_handler_schema
from dagster_cloud.opentelemetry.config.meter_provider import meter_provider_config_schema
from dagster_cloud.opentelemetry.config.metric_reader import metric_reader_config_schema


def opentelemetry_config_schema():
    return {
        "enabled": Field(
            BoolSource,
            description="Enables OpenTelemetry instrumentation.",
            is_required=False,
            default_value=True,
        ),
        "service_name": Field(
            StringSource,
            description="Name of the service to which to send telemetry",
            is_required=False,
            default_value="dagster-cloud-agent",
        ),
        "logging": Field(
            description="The logging configuration.",
            is_required=False,
            config=Shape(
                fields={
                    "enabled": Field(
                        BoolSource,
                        description="Enables logging instrumentation.",
                        is_required=False,
                        default_value=False,
                    ),
                    "exporter": exporter_config_schema("logging"),
                    "processor": log_record_processor_schema(),
                    "handler": logging_handler_schema(),
                },
            ),
        ),
        "metrics": Field(
            description="The metrics configuration.",
            is_required=False,
            config=Shape(
                fields={
                    "enabled": Field(
                        BoolSource,
                        description="Enables metrics instrumentation.",
                        is_required=False,
                        default_value=False,
                    ),
                    "exporter": exporter_config_schema("metrics"),
                    "reader": metric_reader_config_schema(),
                    "provider": meter_provider_config_schema(),
                },
            ),
        ),
        # TODO: Add tracing configuration when it's supported.
        # "tracing": Field(
        #     description="The tracing configuration.",
        #     config=Shape(
        #         fields={
        #             "enabled": Field(
        #                 BoolSource,
        #                 description="Enables tracing instrumentation.",
        #                 is_required=False,
        #                 default_value=False,
        #             ),
        #             "exporter": exporter_config_schema("tracing"),
        #         },
        #     ),
        # ),
    }



import sys
from typing import Any

from .snowflake.dbt_wrapper import dbt_with_snowflake_insights as dbt_with_snowflake_insights
from .snowflake.definitions import (
    create_snowflake_insights_asset_and_schedule as create_snowflake_insights_asset_and_schedule,
)
from .snowflake.snowflake_utils import meter_snowflake_query as meter_snowflake_query

dagster_snowflake_req_imports = {"InsightsSnowflakeResource"}
try:
    from .snowflake.insights_snowflake_resource import (
        InsightsSnowflakeResource as InsightsSnowflakeResource,
    )
except ImportError:
    pass

dagster_bigquery_req_imports = {"InsightsBigQueryResource", "dbt_with_bigquery_insights"}
try:
    from .bigquery.dbt_wrapper import dbt_with_bigquery_insights as dbt_with_bigquery_insights
    from .bigquery.insights_bigquery_resource import (
        InsightsBigQueryResource as InsightsBigQueryResource,
    )
except ImportError:
    pass


# This is overriden in order to provide a better error message
# when the user tries to import a symbol which relies on another integration
# being installed.
def __getattr__(name) -> Any:
    obj = sys.modules[__name__].__dict__.get(name)
    if not obj:
        if name in dagster_snowflake_req_imports:
            raise ImportError(
                f"You are trying to import {name}, but the "
                "dagster-snowflake library is not installed. You can install it with "
                "`pip install dagster-snowflake`.",
            )
        elif name in dagster_bigquery_req_imports:
            raise ImportError(
                f"You are trying to import {name}, but the "
                "dagster-gcp library is not installed. You can install it with "
                "`pip install dagster-gcp`.",
            )
        else:
            raise AttributeError(f"module {__name__} has no attribute {name}")
    return obj



import json
from collections.abc import Iterable, Mapping, Sequence
from datetime import datetime
from typing import Any, Callable, Optional, Union

import dagster._check as check
from dagster._core.errors import (
    DagsterInvariantViolationError,
    DagsterRunAlreadyExists,
    DagsterRunNotFoundError,
    DagsterSnapshotDoesNotExist,
)
from dagster._core.events import DagsterEvent
from dagster._core.execution.backfill import BulkActionsFilter, BulkActionStatus, PartitionBackfill
from dagster._core.execution.telemetry import RunTelemetryData
from dagster._core.remote_representation.origin import RemoteJobOrigin
from dagster._core.snap import ExecutionPlanSnapshot, JobSnap, create_execution_plan_snapshot_id
from dagster._core.storage.dagster_run import (
    DagsterRun,
    JobBucket,
    RunPartitionData,
    RunRecord,
    RunsFilter,
    TagBucket,
)
from dagster._core.storage.runs.base import RunStorage
from dagster._daemon.types import DaemonHeartbeat
from dagster._serdes import (
    ConfigurableClass,
    ConfigurableClassData,
    deserialize_value,
    serialize_value,
)
from dagster._time import datetime_from_timestamp
from dagster._utils.merger import merge_dicts
from dagster_cloud_cli.core.errors import DagsterCloudAgentServerError
from typing_extensions import Self

from .queries import (
    ADD_BACKFILL_MUTATION,
    ADD_DAEMON_HEARTBEAT_MUTATION,
    ADD_EXECUTION_PLAN_SNAPSHOT_MUTATION,
    ADD_HISTORICAL_RUN_MUTATION,
    ADD_PIPELINE_SNAPSHOT_MUTATION,
    ADD_RUN_MUTATION,
    ADD_RUN_TAGS_MUTATION,
    ADD_RUN_TELEMETRY_MUTATION,
    GET_BACKFILL_QUERY,
    GET_BACKFILLS_QUERY,
    GET_DAEMON_HEARTBEATS_QUERY,
    GET_EXECUTION_PLAN_SNAPSHOT_QUERY,
    GET_PIPELINE_SNAPSHOT_QUERY,
    GET_RUN_BY_ID_QUERY,
    GET_RUN_GROUP_QUERY,
    GET_RUN_IDS_QUERY,
    GET_RUN_PARTITION_DATA_QUERY,
    GET_RUN_RECORDS_QUERY,
    GET_RUN_TAG_KEYS_QUERY,
    GET_RUN_TAGS_QUERY,
    GET_RUNS_COUNT_QUERY,
    GET_RUNS_QUERY,
    HAS_EXECUTION_PLAN_SNAPSHOT_QUERY,
    HAS_PIPELINE_SNAPSHOT_QUERY,
    HAS_RUN_QUERY,
    MUTATE_JOB_ORIGIN,
    UPDATE_BACKFILL_MUTATION,
)


def _get_filters_input(filters: Optional[RunsFilter]) -> Optional[dict[str, Any]]:
    filters = check.opt_inst_param(filters, "filters", RunsFilter)

    if filters is None:
        return None

    check.invariant(filters.exclude_subruns is None, "RunsFilter.exclude_subruns is not supported")
    return {
        "runIds": filters.run_ids,
        "pipelineName": filters.job_name,
        "statuses": [status.value for status in filters.statuses],
        "tags": (
            [
                merge_dicts(
                    {"key": tag_key},
                    ({"value": tag_value} if isinstance(tag_value, str) else {"values": tag_value}),
                )
                for tag_key, tag_value in filters.tags.items()
            ]
            if filters.tags
            else []
        ),
        "snapshotId": filters.snapshot_id,
        "updatedAfter": filters.updated_after.timestamp() if filters.updated_after else None,
        "updatedBefore": filters.updated_before.timestamp() if filters.updated_before else None,
        "createdAfter": filters.created_after.timestamp() if filters.created_after else None,
        "createdBefore": filters.created_before.timestamp() if filters.created_before else None,
    }


def _get_bulk_actions_filters_input(
    filters: Optional[BulkActionsFilter],
) -> Optional[dict[str, Any]]:
    filters = check.opt_inst_param(filters, "filters", BulkActionsFilter)
    unsupported_filters = []
    if filters and filters.job_name:
        unsupported_filters.append("job_name")
    if filters and filters.backfill_ids:
        unsupported_filters.append("backfill_ids")
    if filters and filters.tags:
        unsupported_filters.append("tags")

    check.invariant(
        len(unsupported_filters) == 0,
        f"Used the following unsupported filters: {', '.join(unsupported_filters)}.",
    )

    if filters is None:
        return None
    return {
        "statuses": [status.value for status in filters.statuses] if filters.statuses else None,
        "createdAfter": filters.created_after.timestamp() if filters.created_after else None,
        "createdBefore": filters.created_before.timestamp() if filters.created_before else None,
    }


def _run_record_from_graphql(graphene_run_record: dict) -> RunRecord:
    check.dict_param(graphene_run_record, "graphene_run_record")
    return RunRecord(
        storage_id=check.int_elem(graphene_run_record, "storageId"),
        dagster_run=deserialize_value(
            check.str_elem(graphene_run_record, "serializedPipelineRun"),
            DagsterRun,
        ),
        create_timestamp=datetime_from_timestamp(
            check.float_elem(graphene_run_record, "createTimestamp")
        ),
        update_timestamp=datetime_from_timestamp(
            check.float_elem(graphene_run_record, "updateTimestamp")
        ),
        start_time=graphene_run_record.get("startTime"),
        end_time=graphene_run_record.get("endTime"),
    )


def _get_bucket_input(bucket_by: Optional[Union[JobBucket, TagBucket]]) -> Optional[dict[str, Any]]:
    if not bucket_by:
        return None

    bucket_type = "job" if isinstance(bucket_by, JobBucket) else "tag"
    tag_key = bucket_by.tag_key if isinstance(bucket_by, TagBucket) else None
    values = bucket_by.job_names if isinstance(bucket_by, JobBucket) else bucket_by.tag_values
    return {
        "bucketType": bucket_type,
        "tagKey": tag_key,
        "values": values,
        "bucketLimit": bucket_by.bucket_limit,
    }


class GraphQLRunStorage(RunStorage, ConfigurableClass):
    def __init__(
        self, inst_data: Optional[ConfigurableClassData] = None, override_graphql_client=None
    ):
        """Initialize this class directly only for test (using `override_graphql_client`).
        Use the ConfigurableClass machinery to init from instance yaml.
        """
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        self._override_graphql_client = override_graphql_client

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @classmethod
    def from_config_value(cls, inst_data: ConfigurableClassData, config_value: Any) -> Self:
        return cls(inst_data=inst_data)

    @property
    def _graphql_client(self):
        return (
            self._override_graphql_client
            if self._override_graphql_client
            else self._instance.graphql_client  # pyright: ignore[reportAttributeAccessIssue]
        )

    def _execute_query(self, query, variables=None, idempotent_mutation=False):
        return self._graphql_client.execute(
            query, variable_values=variables, idempotent_mutation=idempotent_mutation
        )

    def add_historical_run(
        self, dagster_run: DagsterRun, run_creation_time: datetime
    ) -> DagsterRun:
        check.inst_param(dagster_run, "dagster_run", DagsterRun)
        res = self._execute_query(
            ADD_HISTORICAL_RUN_MUTATION,
            variables={
                "serializedPipelineRun": serialize_value(dagster_run),
                "runCreationTime": run_creation_time.timestamp(),
            },
        )
        result = res["data"]["runs"]["addHistoricalRun"]
        error = result.get("error")
        # Special-case some errors to match the RunStorage API
        if error:
            if error["className"] == "DagsterRunAlreadyExists":
                raise DagsterRunAlreadyExists(error["message"])
            if error["className"] == "DagsterSnapshotDoesNotExist":
                raise DagsterSnapshotDoesNotExist(error["message"])
            else:
                raise DagsterCloudAgentServerError(res)
        return dagster_run

    def add_run(self, dagster_run: DagsterRun):
        check.inst_param(dagster_run, "dagster_run", DagsterRun)
        res = self._execute_query(
            ADD_RUN_MUTATION,
            variables={"serializedPipelineRun": serialize_value(dagster_run)},
        )

        result = res["data"]["runs"]["addRun"]
        error = result.get("error")

        # Special-case some errors to match the RunStorage API
        if error:
            if error["className"] == "DagsterRunAlreadyExists":
                raise DagsterRunAlreadyExists(error["message"])
            if error["className"] == "DagsterSnapshotDoesNotExist":
                raise DagsterSnapshotDoesNotExist(error["message"])
            else:
                raise DagsterCloudAgentServerError(res)

        return dagster_run

    def handle_run_event(
        self, run_id: str, event: DagsterEvent, update_timestamp: Optional[datetime] = None
    ):
        raise NotImplementedError("Should never be called by an agent client")

    @property
    def supports_bucket_queries(self) -> bool:
        return False

    def get_runs(  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        self,
        filters: Optional[RunsFilter] = None,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
        bucket_by: Optional[Union[JobBucket, TagBucket]] = None,
        ascending: bool = False,
    ) -> Iterable[DagsterRun]:
        res = self._execute_query(
            GET_RUNS_QUERY,
            variables={
                "filters": _get_filters_input(filters),
                "cursor": check.opt_str_param(cursor, "cursor"),
                "limit": check.opt_int_param(limit, "limit"),
                "bucketBy": _get_bucket_input(bucket_by),
                "ascending": check.bool_param(ascending, "ascending"),
            },
        )
        return [deserialize_value(run, DagsterRun) for run in res["data"]["runs"]["getRuns"]]

    def get_run_ids(
        self,
        filters: Optional[RunsFilter] = None,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> Sequence[str]:
        res = self._execute_query(
            GET_RUN_IDS_QUERY,
            variables={
                "filters": _get_filters_input(filters),
                "cursor": check.opt_str_param(cursor, "cursor"),
                "limit": check.opt_int_param(limit, "limit"),
            },
        )
        return res["data"]["runs"]["getRunIds"]

    def get_runs_count(self, filters: Optional[RunsFilter] = None) -> int:
        res = self._execute_query(
            GET_RUNS_COUNT_QUERY,
            variables={
                "filters": _get_filters_input(filters),
            },
        )
        return res["data"]["runs"]["getRunsCount"]

    def get_run_group(self, run_id: str) -> Optional[tuple[str, Iterable[DagsterRun]]]:  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        res = self._execute_query(
            GET_RUN_GROUP_QUERY, variables={"runId": check.str_param(run_id, "run_id")}
        )
        run_group_or_error = res["data"]["runs"]["getRunGroupOrError"]
        if run_group_or_error["__typename"] == "SerializedRunGroup":
            return (
                run_group_or_error["rootRunId"],
                [
                    deserialize_value(run, DagsterRun)
                    for run in run_group_or_error["serializedRuns"]
                ],
            )
        elif run_group_or_error["__typename"] == "RunNotFoundError":
            raise DagsterRunNotFoundError(invalid_run_id=run_group_or_error["runId"])
        else:
            raise DagsterInvariantViolationError(f"Unexpected getRunGroupOrError response {res}")

    def get_run_by_id(self, run_id: str) -> Optional[DagsterRun]:
        res = self._execute_query(
            GET_RUN_BY_ID_QUERY, variables={"runId": check.str_param(run_id, "run_id")}
        )
        run = res["data"]["runs"]["getRunById"]
        if run is None:
            return None

        return deserialize_value(run, DagsterRun)

    def get_run_records(
        self,
        filters: Optional[RunsFilter] = None,
        limit: Optional[int] = None,
        order_by: Optional[str] = None,
        ascending: bool = False,
        cursor: Optional[str] = None,
        bucket_by: Optional[Union[JobBucket, TagBucket]] = None,
    ) -> list[RunRecord]:
        res = self._execute_query(
            GET_RUN_RECORDS_QUERY,
            variables={
                "filters": _get_filters_input(filters),
                "limit": check.opt_int_param(limit, "limit"),
                "orderBy": check.opt_str_param(order_by, "order_by"),
                "ascending": check.opt_bool_param(ascending, "ascending"),
                "cursor": check.opt_str_param(cursor, "cursor"),
                "bucketBy": _get_bucket_input(bucket_by),
            },
        )
        return [_run_record_from_graphql(record) for record in res["data"]["runs"]["getRunRecords"]]

    def get_run_tags(
        self,
        tag_keys: Sequence[str],
        value_prefix: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> Sequence[tuple[str, set[str]]]:
        res = self._execute_query(
            GET_RUN_TAGS_QUERY,
            variables={
                "jsonTagKeys": (json.dumps(check.list_param(tag_keys, "tag_keys", of_type=str))),
                "valuePrefix": check.opt_str_param(value_prefix, "value_prefix"),
                "limit": check.opt_int_param(limit, "limit"),
            },
        )
        return [
            (run_tag["key"], set(run_tag["values"]))
            for run_tag in res["data"]["runs"]["getRunTags"]
        ]

    def get_run_tag_keys(self) -> Sequence[str]:
        res = self._execute_query(GET_RUN_TAG_KEYS_QUERY)
        return [run_tag_key for run_tag_key in res["data"]["runs"]["getRunTagKeys"]]

    def add_run_tags(self, run_id: str, new_tags: Mapping[str, str]):
        self._execute_query(
            ADD_RUN_TAGS_MUTATION,
            variables={
                "runId": check.str_param(run_id, "run_id"),
                "jsonNewTags": json.dumps(
                    check.mapping_param(new_tags, "new_tags", key_type=str, value_type=str)
                ),
            },
            idempotent_mutation=True,
        )

    def has_run(self, run_id: str) -> bool:
        res = self._execute_query(
            HAS_RUN_QUERY, variables={"runId": check.str_param(run_id, "run_id")}
        )
        return res["data"]["runs"]["hasRun"]

    def has_job_snapshot(self, pipeline_snapshot_id: str) -> bool:  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        res = self._execute_query(
            HAS_PIPELINE_SNAPSHOT_QUERY,
            variables={
                "pipelineSnapshotId": check.str_param(pipeline_snapshot_id, "pipeline_snapshot_id")
            },
        )
        return res["data"]["runs"]["hasPipelineSnapshot"]

    def add_job_snapshot(  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        self, pipeline_snapshot: JobSnap, snapshot_id: Optional[str] = None
    ) -> str:
        self._execute_query(
            ADD_PIPELINE_SNAPSHOT_MUTATION,
            variables={
                "serializedPipelineSnapshot": serialize_value(
                    check.inst_param(pipeline_snapshot, "pipeline_snapshot", JobSnap)
                ),
                "snapshotId": snapshot_id,
            },
        )
        return snapshot_id if snapshot_id else pipeline_snapshot.snapshot_id

    def get_job_snapshot(self, pipeline_snapshot_id: str) -> JobSnap:  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        res = self._execute_query(
            GET_PIPELINE_SNAPSHOT_QUERY,
            variables={
                "pipelineSnapshotId": check.str_param(pipeline_snapshot_id, "pipeline_snapshot_id")
            },
        )
        return deserialize_value(res["data"]["runs"]["getPipelineSnapshot"], JobSnap)

    def has_execution_plan_snapshot(self, execution_plan_snapshot_id: str) -> bool:
        res = self._execute_query(
            HAS_EXECUTION_PLAN_SNAPSHOT_QUERY,
            variables={
                "executionPlanSnapshotId": check.str_param(
                    execution_plan_snapshot_id, "execution_plan_snapshot_id"
                )
            },
        )
        return res["data"]["runs"]["hasExecutionPlanSnapshot"]

    def add_execution_plan_snapshot(
        self, execution_plan_snapshot: ExecutionPlanSnapshot, snapshot_id: Optional[str] = None
    ) -> str:
        self._execute_query(
            ADD_EXECUTION_PLAN_SNAPSHOT_MUTATION,
            variables={
                "serializedExecutionPlanSnapshot": serialize_value(
                    check.inst_param(
                        execution_plan_snapshot, "execution_plan_snapshot", ExecutionPlanSnapshot
                    )
                ),
                "snapshotId": snapshot_id,
            },
        )
        return (
            snapshot_id
            if snapshot_id
            else create_execution_plan_snapshot_id(execution_plan_snapshot)
        )

    def get_execution_plan_snapshot(self, execution_plan_snapshot_id: str) -> ExecutionPlanSnapshot:
        res = self._execute_query(
            GET_EXECUTION_PLAN_SNAPSHOT_QUERY,
            variables={
                "executionPlanSnapshotId": check.str_param(
                    execution_plan_snapshot_id, "execution_plan_snapshot_id"
                )
            },
        )
        return deserialize_value(
            res["data"]["runs"]["getExecutionPlanSnapshot"], ExecutionPlanSnapshot
        )

    def get_run_partition_data(self, runs_filter: RunsFilter) -> list[RunPartitionData]:
        res = self._execute_query(
            GET_RUN_PARTITION_DATA_QUERY,
            variables={
                "runsFilter": _get_filters_input(runs_filter),
            },
        )
        return [
            deserialize_value(result, RunPartitionData)
            for result in res["data"]["runs"]["getRunPartitionData"]
        ]

    def add_daemon_heartbeat(self, daemon_heartbeat: DaemonHeartbeat):
        self._execute_query(
            ADD_DAEMON_HEARTBEAT_MUTATION,
            variables={
                "serializedDaemonHeartbeat": serialize_value(
                    check.inst_param(daemon_heartbeat, "daemon_heartbeat", DaemonHeartbeat)
                )
            },
        )

    def supports_run_telemetry(self) -> bool:
        return True

    def add_run_telemetry(
        self,
        run_telemetry: RunTelemetryData,
        tags: Optional[dict[str, str]] = None,
    ) -> None:
        if tags is None:
            tags = {}

        self._execute_query(
            ADD_RUN_TELEMETRY_MUTATION,
            variables={
                "serializedTelemetry": serialize_value(
                    check.inst_param(run_telemetry, "run_telemetry", RunTelemetryData)
                ),
                "serializedTags": serialize_value(
                    check.dict_param(tags, "tags", key_type=str, value_type=str)
                ),
            },
        )

    def build_missing_indexes(
        self, print_fn: Callable = lambda _: None, force_rebuild_all: bool = False
    ):
        raise Exception("Not allowed to build indexes from user cloud")

    def get_daemon_heartbeats(self) -> dict[str, DaemonHeartbeat]:
        res = self._execute_query(GET_DAEMON_HEARTBEATS_QUERY)
        return {
            key: deserialize_value(heartbeat, DaemonHeartbeat)
            for key, heartbeat in json.loads(res["data"]["runs"]["getDaemonHeartbeats"]).items()
        }

    def delete_run(self, run_id: str):
        raise Exception("Not allowed to delete runs from user cloud")

    def wipe_daemon_heartbeats(self):
        raise Exception("Not allowed to wipe heartbeats from user cloud")

    def wipe(self):
        raise Exception("Not allowed to wipe runs from user cloud")

    def has_bulk_actions_table(self) -> bool:
        return False

    def get_backfills(
        self,
        filters: Optional[BulkActionsFilter] = None,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
        status: Optional[BulkActionStatus] = None,
    ):
        """Get a list of partition backfills."""
        res = self._execute_query(
            GET_BACKFILLS_QUERY,
            variables={
                "status": status.value if status else None,
                "cursor": check.opt_str_param(cursor, "cursor"),
                "limit": check.opt_int_param(limit, "limit"),
                "filters": _get_bulk_actions_filters_input(filters),
            },
        )
        return [
            deserialize_value(backfill, PartitionBackfill)
            for backfill in res["data"]["runs"]["getBackfills"]
        ]

    def get_backfills_count(self, filters: Optional[BulkActionsFilter] = None) -> int:
        raise NotImplementedError("get_backfills_count is not callable from user cloud.")

    def get_backfill(self, backfill_id: str) -> PartitionBackfill:
        """Get a single partition backfill."""
        res = self._execute_query(GET_BACKFILL_QUERY, variables={"backfillId": backfill_id})
        backfill = res["data"]["runs"]["getBackfill"]
        return deserialize_value(backfill, PartitionBackfill)

    def add_backfill(self, partition_backfill: PartitionBackfill):
        """Add partition backfill to run storage."""
        self._execute_query(
            ADD_BACKFILL_MUTATION,
            variables={"serializedPartitionBackfill": serialize_value(partition_backfill)},
        )

    def update_backfill(self, partition_backfill: PartitionBackfill):
        """Update a partition backfill in run storage."""
        self._execute_query(
            UPDATE_BACKFILL_MUTATION,
            variables={"serializedPartitionBackfill": serialize_value(partition_backfill)},
        )

    def get_cursor_values(self, keys: set[str]):  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        return NotImplementedError("KVS is not supported from the user cloud")

    def set_cursor_values(self, pairs: Mapping[str, str]):  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        return NotImplementedError("KVS is not supported from the user cloud")

    # Migrating run history
    def replace_job_origin(self, run: DagsterRun, job_origin: RemoteJobOrigin):
        self._execute_query(
            MUTATE_JOB_ORIGIN,
            variables={
                "runId": run.run_id,
                "serializedJobOrigin": serialize_value(job_origin),
            },
        )



from dagster_cloud_cli.core.headers.auth import (
    API_TOKEN_HEADER as API_TOKEN_HEADER,
    AUTHORIZATION_HEADER as AUTHORIZATION_HEADER,
    BEARER_HEADER_SCHEME as BEARER_HEADER_SCHEME,
    DEPLOYMENT_NAME_HEADER as DEPLOYMENT_NAME_HEADER,
)



import logging
import os
from collections.abc import Generator
from concurrent.futures import Future, TimeoutError
from contextlib import contextmanager
from queue import Empty, Full, Queue
from threading import Lock
from typing import Callable, Generic, Optional, TypeVar

import dagster._check as check

from dagster_cloud.instrumentation import Instrumentation, NoOpInstrumentation

logger = logging.getLogger(__name__)

I = TypeVar("I")  # noqa: E741
O = TypeVar("O")  # noqa: E741
QueueItem = tuple[I, Future[O]]


DEFAULT_MAX_WAIT_MS = 1000
DEFAULT_MAX_BATCH_SIZE = 100
DEFAULT_MAX_QUEUE_SIZE = 1000


def _get_override_for_name(setting: str, name: str) -> Optional[int]:
    env_name = f"DAGSTER_BATCHING__{name.upper().replace('-', '_')}__{setting.upper()}"
    value = os.getenv(env_name)
    if value is None:
        return None

    try:
        value_int = int(value)
        if value_int <= 0:
            logger.warning(
                f"Environment variable misconfiguration for {env_name} (should be positive int, got: '{value}')"
            )
            return None
        return value_int
    except ValueError:
        logger.warning(
            f"Environment variable misconfiguration for {env_name} (should be positive int, got: '{value}')"
        )
        return None


def _get_config(
    setting: str, name: str, passed_in_default: Optional[int], global_default: int
) -> int:
    override = _get_override_for_name(setting, name)
    if override is not None:
        return override

    if passed_in_default is not None:
        return passed_in_default

    return global_default


class Batcher(Generic[I, O]):
    """the basic algorithm is.

    1. insert (item, future) into queue
    2. wait for future to complete, with max timeout
      2a. if future completes, return result
      2b. on timeout, acquire lock, then drain the queue until
          the future completes

    NOTE: if the queue is full, submit() will raise an exception
    NOTE: the lock means that only one thread will ever be running the batcher_fn
          at a time. the algorithm would still be correct without the lock but
          locking leads to larger batches. HOWEVER without the lock we might try
          to submit empty batches, which there is currently an invariant to protect
          against
    NOTE: the max queue size is meant to cap the number of inflight requests
          in order to fail faster if the underlying function is taking too long
          (database issues).

    Configuration for queue size, max wait, and batch size is specified (by priority order) by:

    1. an env var override (of the form DAGSTER_BATCHING__TEST__MAX_WAIT_MS -- see _get_override_for_name)
    2. the passed in value
    3. the default (specified in this file)
    """

    def __init__(
        self,
        name: str,
        batcher_fn: Callable[[list[I]], list[O]],
        max_queue_size: Optional[int] = None,
        max_batch_size: Optional[int] = None,
        max_wait_ms: Optional[int] = None,
        instrumentation: Optional[Instrumentation] = None,
    ) -> None:
        check.invariant(
            max_wait_ms is None or max_wait_ms > 0,
            "max wait, if provided, must be set to a positive integer",
        )
        check.invariant(
            max_queue_size is None or max_queue_size > 0,
            "max queue size, if provided, must be set to a positive integer",
        )
        check.invariant(
            max_batch_size is None or max_batch_size > 0,
            "max batch size, if provided, must be set to a positive integer",
        )
        if max_queue_size and max_batch_size:
            check.invariant(
                max_batch_size <= max_queue_size,
                "if max batch size and max queue size are provided, max batch size must be "
                "less than or equal to max queue size",
            )
        self._name = name
        self._batcher_fn = batcher_fn
        self._max_batch_size = _get_config(
            "max_batch_size", name, max_batch_size, DEFAULT_MAX_BATCH_SIZE
        )
        self._max_wait_ms: float = _get_config(
            "max_wait_ms", name, max_wait_ms, DEFAULT_MAX_WAIT_MS
        )
        config_max_queue_size = _get_config(
            "max_queue_size", name, max_queue_size, DEFAULT_MAX_QUEUE_SIZE
        )
        self._queue: Queue[QueueItem] = Queue(maxsize=config_max_queue_size)
        self._drain_lock = Lock()
        self._instrumentation = (instrumentation or NoOpInstrumentation()).tags([f"batcher:{name}"])

    def _submit_batch(self, batch: list[QueueItem]) -> None:
        check.invariant(len(batch) > 0, "should never submit an empty batch")
        self._instrument_batch_size(len(batch))
        try:
            with self._time("batcher_fn"):
                results = self._batcher_fn([i for i, _ in batch])
        except Exception as e:
            for _, fut in batch:
                fut.set_exception(e)
        else:
            check.invariant(
                len(results) == len(batch), "batcher returned fewer results than expected"
            )
            for (_, fut), result in zip(batch, results):
                fut.set_result(result)

    def _build_batch(self) -> list[QueueItem]:
        batch = []
        for _ in range(self._max_batch_size):
            try:
                batch.append(self._queue.get(block=False))
            except Empty:
                break
        return batch

    @contextmanager
    def _lock(self) -> Generator[None, None, None]:
        with self._time("lock_acquisition"):
            self._drain_lock.acquire()
        try:
            yield
        finally:
            self._drain_lock.release()

    def _drain_batch(self, fut: Future[O]) -> O:
        with self._lock(), self._time("drain_batch"):
            while not fut.done():
                self._submit_batch(self._build_batch())
            return fut.result()

    def submit(self, i: I) -> O:
        with self._time("submit"):
            fut: Future[O] = Future()
            try:
                self._queue.put((i, fut), block=False)
            except Full:
                self._instrumentation.increment("dagster.batching.full")
                logger.exception(f"Batching queue for batcher {self._name} is full!")
                raise
            else:
                try:
                    queue_size = self._queue.qsize()
                    self._instrument_queue_size(queue_size)
                    timeout = 0 if queue_size >= self._max_batch_size else self._max_wait_ms / 1000
                    return fut.result(timeout=timeout)
                except TimeoutError:
                    self._instrumentation.increment("dagster.batching.timeout")
                    self._drain_batch(fut)
                    return fut.result()

    def _instrument_queue_size(self, queue_size: int) -> None:
        self._instrumentation.histogram("dagster.batching.queue_size", queue_size)
        for bucket in [5, 10, 100]:
            if queue_size >= bucket:
                self._instrumentation.increment(f"dagster.batching.queue_size.ge_{bucket}")
            else:
                break

    def _instrument_batch_size(self, batch_size: int) -> None:
        self._instrumentation.histogram("dagster.batching.batch_size", batch_size)
        for bucket in [5, 10, 100]:
            if batch_size >= bucket:
                self._instrumentation.increment(f"dagster.batching.batch_size.ge_{bucket}")
            else:
                break

    @contextmanager
    def _time(self, metric_name: str) -> Generator[None, None, None]:
        with self._instrumentation.instrument_context(
            f"dagster.batching.{metric_name}",
            buckets_ms=[10, 100, 500, 1000],
        ):
            yield
